from typing import (
    Collection,
    Dict,
    List,
    NamedTuple,
    NewType,
    Optional,
    Sequence,
    TypeVar,
)

from sr.comp.types import ArenaName, MatchNumber, TLA

T = TypeVar('T')
ID = NewType('ID', str)
RawMatch = Dict[ArenaName, List[Optional[TLA]]]


class BadMatch(NamedTuple):
    arena: ArenaName
    num: MatchNumber
    num_teams: int


class Configuration(NamedTuple):
    arena_ids: Collection[ArenaName]
    team_ids: Sequence[TLA]

    teams_per_game: int

    first_match_number: MatchNumber

    @property
    def num_arenas(self) -> int:
        return len(self.arena_ids)

    @property
    def num_teams(self) -> int:
        return len(self.team_ids)



import time
from datetime import timedelta
from enum import Enum

from sr.comp.comp import SRComp

__description__ = "A way of controlling lights at the competition."


class State(Enum):
    idle = 0
    match = 1
    match_ending = 2
    pre_match = 3


class CompetitionStateMachine:
    def __init__(self, comp):
        self.comp = comp

    @property
    def now(self):
        return self.comp.schedule.datetime_now

    def wait_until(self, date):
        while True:
            delta = date - self.now
            s = delta.total_seconds()
            if s <= 0:
                break
            time.sleep(s)

    @property
    def current_match(self):
        current_matches = list(self.comp.schedule.matches_at(self.now))
        if current_matches:
            return current_matches[0]
        else:
            return None

    @property
    def next_match(self):
        # assumes list is in time order
        now = self.now
        for slot in self.comp.schedule.matches:
            for match in slot.values():
                if self.get_game_end_time(match) >= now:
                    return match
        return None

    def get_game_start_time(self, match):
        match_slot_lengths = self.comp.schedule.match_slot_lengths
        return match.start_time + match_slot_lengths['pre']

    def get_game_end_time(self, match):
        match_slot_lengths = self.comp.schedule.match_slot_lengths
        return match.end_time - match_slot_lengths['post']

    def is_match_in_game(self, match):
        a = self.get_game_start_time(match)
        b = self.get_game_end_time(match)
        return a <= self.now <= b

    @property
    def current_state(self):
        current_match = self.current_match

        if current_match is not None and self.is_match_in_game(current_match):
            game_end_time = self.get_game_end_time(current_match)
            game_ending_time = game_end_time - timedelta(seconds=10)
            if self.now <= game_ending_time:
                return State.match, game_ending_time
            if self.now <= game_end_time:
                return State.match_ending, game_end_time

        next_match = self.next_match

        if next_match is None:
            return State.idle, self.now + timedelta(days=100)

        game_start_time = self.get_game_start_time(next_match)
        game_starting_time = game_start_time - timedelta(seconds=5)
        if game_starting_time <= self.now <= game_start_time:
            return State.pre_match, game_start_time

        return State.idle, game_starting_time

    def transition(self):
        prev_state = None
        while True:
            state, transition_date = self.current_state
            yield prev_state, state, transition_date
            self.wait_until(transition_date)
            prev_state = state


class LightingController:
    def __init__(self, comp):
        self.comp = comp
        self.state = CompetitionStateMachine(comp)

    def transition(self, prev_state, state, transition_date):
        pass

    def run(self):
        for prev_state, state, transition_date in self.state.transition():
            self.transition(prev_state, state, transition_date)
            print(state, "until", transition_date)


class MidiLightingController(LightingController):
    tracks = {
        State.idle: 0,
        State.match: 1,
        State.pre_match: 2,
        State.match_ending: 3,
    }

    def __init__(self, comp, midi):
        super().__init__(comp)

        self.midi = midi
        self.cur_playback = None

    def note_on(self, note, velocity):
        import mido  # type: ignore[import]
        msg = mido.Message('note_on', note=note, velocity=velocity)
        self.midi.send(msg)

    def start_playback(self, note, velocity=127):
        self.note_on(note, velocity)

    def stop_playback(self, note):
        self.note_on(note, 0)

    def transition(self, prev_state, state, transition_date):
        if self.cur_playback is not None:
            self.stop_playback(self.cur_playback)

        playback = self.tracks[state]
        self.start_playback(playback)
        self.cur_playback = playback


def command(args):
    comp = SRComp(args.compstate)

    controller = None

    if args.dry_run:
        controller = LightingController(comp)
    elif args.midi:
        import mido
        output = mido.open_output(args.midi)
        controller = MidiLightingController(comp, output)

    assert controller

    controller.run()


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'lighting-controller',
        help=__description__,
        description=__description__,
    )
    parser.add_argument('compstate', help="Competition state repository.")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--midi', help="The MIDI port to use.")
    group.add_argument('--dry-run', action='store_true', help="Dry run.")

    parser.set_defaults(func=command)



import random
import sys
from collections import Counter
from itertools import product
from math import gcd


class PatienceCounter:
    def __init__(self, threshold):
        self.threshold = threshold
        self.level = 0

    def bump(self):
        self.level += 1

    def reset(self):
        self.level = 0

    def reached(self):
        return self.level >= self.threshold


def prime_factors(n):
    d = 2
    while d * d <= n:
        while n % d == 0:
            yield d
            n //= d
        d += 1
    if n > 1:
        yield n


class Scheduler:
    def __init__(
        self,
        teams,
        max_match_periods,
        arenas=('main',),
        num_corners=4,
        random=random,
        appearances_per_round=1,
        separation=2,
        max_matchups=2,
        enable_lcg=True,
        base_matches=(),
    ):
        self.tag = ''
        self.num_corners = num_corners
        self.random = random
        self.arenas = tuple(arenas)
        self.max_match_periods = max_match_periods
        self.appearances_per_round = appearances_per_round
        self._base_matches = list(base_matches)
        for match in base_matches:
            for n, entry in enumerate(match):
                if entry is None:
                    match[n] = '~'
        self._calculate_teams(teams)
        self._calculate_rounds()
        if len(self._base_matches) % self.round_length > 0:
            self.lprint(
                "Warning: matches for partial reschedule are not a multiple of "
                "the round-length",
            )
        self.separation = separation
        self.max_matchups = max_matchups
        if enable_lcg:
            self._compute_lcg_params()
        else:
            self._lcg_params = None

    def lprint(self, *args, **kwargs):
        if self.tag:
            print(self.tag, end='', file=sys.stderr)
        print(*args, file=sys.stderr, **kwargs)

    @property
    def entrants_per_match_period(self):
        return len(self.arenas) * self.num_corners

    def _is_pseudo(self, team):
        return team[0] == '~'

    def _calculate_teams(self, base_teams):
        teams = list(base_teams) * self.appearances_per_round
        # account for overflow
        overflow = (
            self.entrants_per_match_period -
            (len(teams) % self.entrants_per_match_period)
        )
        if overflow < self.entrants_per_match_period:
            for n in range(overflow):
                teams.append(f'~{n}')
        self._teams = teams

    @property
    def total_matches(self):
        return self.num_rounds * self.round_length

    def _calculate_rounds(self):
        self.num_rounds = int(
            self.max_match_periods *
            self.entrants_per_match_period //
            len(self._teams),
        )
        self.round_length = len(self._teams) // self.entrants_per_match_period

    def _validate(
        self,
        schedule,
        matchup_max=None,
        matchup_impatience_bump=lambda: None,
    ):
        is_pseudo = self._is_pseudo
        if matchup_max is None:
            matchup_max = self.max_matchups
        multi_per_match_mode = self.appearances_per_round > 1
        # 4 tests in this function:
        #  (1) validate that teams aren't scheduled too tightly
        #  (2) validate that matchups aren't too frequent
        #  (3) validate that no match has two teams sitting out (or if it is, that it's blank)
        # if operating multiple appearances per match, also:
        #  (4) make sure that a team doesn't appear in a match twice
        matchups = Counter()
        for match_id, match in enumerate(schedule):
            entrants = set(
                entrant
                for entrant in match
                if not is_pseudo(entrant)
            )
            if multi_per_match_mode:
                # Test constraint (4)
                if len(entrants) != len([
                    entrant
                    for entrant in match
                    if not is_pseudo(entrant)
                ]):
                    return False
            # Test constraint (1)
            previous_matches = schedule[match_id - self.separation:match_id]
            for previous_match in previous_matches:
                for previous_entrant in previous_match:
                    if is_pseudo(previous_entrant):
                        continue
                    if previous_entrant in entrants:
                        return False
            # Update constraint (2)
            for arena_id in range(len(self.arenas)):
                game = match[arena_id * self.num_corners:(arena_id + 1) * self.num_corners]
                for a, b in product(game, repeat=2):
                    if a >= b:
                        continue
                    a_pseudo, b_pseudo = is_pseudo(a), is_pseudo(b)
                    # Check constraint (3) while we're here
                    if (
                        a_pseudo and
                        b_pseudo and
                        not all(is_pseudo(x) for x in game)
                    ):
                        return False
                    elif not a_pseudo and not b_pseudo:
                        matchups.update([(a, b)])
        # No collisions, determine whether teams face a broad range of other teams
        top_repeated_matchups = max(matchups.values())
        if top_repeated_matchups > self.max_matchups:
            # team faces off against one other team too many times
            matchup_impatience_bump()
            return False
        # No objections, your honour!
        return True

    def _compute_lcg_params(self):
        m = len(self._teams)
        for a in range(m - 1, 1, -1):
            am1 = a - 1
            if am1 % 4 != 0:
                continue
            if any(am1 % factor != 0 for factor in prime_factors(m)):
                continue
            for c in range(m - 1, 0, -1):
                if gcd(c, m) != 1:
                    continue
                epm = self.entrants_per_match_period
                acceptable = True
                for sm in range(1, self.separation + 1):
                    overlap = 1 + self.separation - sm
                    dst_a = 0
                    dst_b = epm * overlap
                    src_a = (self.round_length - sm) * epm
                    src_b = (1 + self.round_length - sm) * epm
                    src = set((a * x + c) % m for x in range(src_a, src_b))
                    dst = set(range(dst_a, dst_b))
                    if not src.isdisjoint(dst):
                        acceptable = False
                if not acceptable:
                    continue
                self._lcg_params = (a, c)
                self.lprint(f"Found LCG settings: ({a}, {c})")
                return
        self.lprint("No valid LCG parameters")
        self._lcg_params = None

    def _lcg_permute(self, teams):
        if self._lcg_params is None:
            return None
        a, c = self._lcg_params
        m = len(teams)
        if m != len(self._teams):
            return None
        permutation = [teams[(a * n + c) % m] for n in range(len(teams))]
        if set(permutation) != set(teams):
            raise ValueError("permutation fault")
        return permutation

    def run(self):
        matchup_impatience = PatienceCounter(200000)
        max_matchups = self.max_matchups
        matches = list(self._base_matches)
        teams = list(self._teams)
        self.random.shuffle(teams)
        while (
            len(matches) < self.total_matches and
            len(matches) + self.round_length <= self.max_match_periods
        ):
            this_round = len(matches) // self.round_length
            self.lprint("Scheduling round {round} ({prev}/{tot} complete)".format(
                round=this_round,
                prev=len(matches),
                tot=self.total_matches,
            ))
            # Attempt the LCG
            lcg_round = self._lcg_permute(teams)
            if lcg_round is not None:
                matches_prime = matches + self._match_partition(lcg_round)
                if self._validate(matches_prime, max_matchups, matchup_impatience.bump):
                    matches = matches_prime
                    self.lprint("  completed via LCG permutation")
                    continue
            for _ in range(10000):
                if matchup_impatience.reached():
                    matchup_impatience.reset()
                    self.lprint("  Easing off on matchup constraint.")
                    max_matchups += 1
                self.random.shuffle(teams)
                matches_prime = matches + self._match_partition(teams)
                if self._validate(matches_prime, max_matchups, matchup_impatience.bump):
                    matches = matches_prime
                    break
            else:
                if len(matches) > len(self._base_matches):
                    self.lprint("  backtracking")
                    matches = matches[:-self.round_length]
        return self._clean(matches)

    def _match_partition(self, teams):
        entries = []
        for n in range(0, len(teams), self.entrants_per_match_period):
            entries.append(teams[n:n + self.entrants_per_match_period])
        return entries

    def _clean(self, matches):
        def get_match(match_id, match):
            data = {}
            for arena_id, arena in enumerate(self.arenas):
                entrants = match[arena_id * self.num_corners:(arena_id + 1) * self.num_corners]
                # Shuffle entrants to get statistically sensible zone distribution
                if match_id >= len(self._base_matches):  # don't shuffle provided matches!
                    self.random.shuffle(entrants)
                entrants = [
                    None if self._is_pseudo(entrant) else entrant
                    for entrant in entrants
                ]
                data[arena] = entrants
            return data
        return {match_id: get_match(match_id, match) for match_id, match in enumerate(matches)}



import os
import sys


sys.path.insert(0, os.path.abspath('..'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon'
]

source_suffix = '.rst'
master_doc = 'index'

project = 'sr.comp.cli'
copyright = '2015, Student Robotics'

release = '1.0.0'
version = '1.0.0'

html_theme = 'alabaster'

intersphinx_mapping = {'http://docs.python.org/': None}



from pathlib import Path
from typing import Dict, Iterable, List, Tuple

from sr.comp.types import ArenaName, MatchNumber, TLA

from .types import ID, RawMatch


def parse_ids(ids: str, sep: str = ',') -> List[ID]:
    return [ID(x) for x in ids.split(sep)]


def tidy(lines: Iterable[str]) -> List[str]:
    "Strip comments and trailing whitespace"
    schedule = []
    for line in lines:
        idx = line.find('#')
        if idx > -1:
            line = line[:idx]

        line = line.strip()

        if line:
            schedule.append(line)

    return schedule


def league_yaml_path(compstate_path: Path) -> Path:
    return compstate_path / 'league.yaml'


def dump_league_yaml(
    matches: Dict[MatchNumber, RawMatch],
    file_path: Path,
) -> None:
    import yaml

    with open(file_path, 'w') as lfp:
        empty = {'matches': matches}
        yaml.dump(empty, lfp)


def load_league_yaml(league_yaml: Path) -> Dict[MatchNumber, RawMatch]:
    import yaml

    with open(league_yaml) as lfp:
        data = yaml.load(lfp)
        matches: Dict[MatchNumber, RawMatch] = data['matches']
        return matches


def load_teams_areans(compstate_path: Path) -> Tuple[List[TLA], List[ArenaName], int]:
    from sr.comp import arenas, teams

    team_ids = sorted(teams.load_teams(compstate_path / 'teams.yaml').keys())
    arenas_yaml = compstate_path / 'arenas.yaml'
    arena_ids = sorted(arenas.load_arenas(arenas_yaml).keys())
    num_corners = len(arenas.load_corners(arenas_yaml))

    return team_ids, arena_ids, num_corners


def load_ids_schedule(
    schedule_lines: Iterable[str],
    num_arenas: int,
    teams_per_game: int,
) -> Tuple[List[ID], List[List[ID]]]:
    """
    Converts an iterable of strings containing pipe-separated ids into
    a tuple: ``(ids, schedule)``. The ``ids`` is a list of unique ids
    in the order which they first appear, the ``schedule`` is a list of
    lists of ids in each line.
    """

    max_teams_per_slot = teams_per_game * num_arenas

    ids: List[ID] = []
    schedule: List[List[ID]] = []

    for match_num, match in enumerate(schedule_lines):
        match_ids = parse_ids(match, sep='|')

        uniq_match_ids = set(match_ids)
        if len(match_ids) != len(uniq_match_ids):
            raise ValueError(
                f"Match {match_num} contains the same id more than once. "
                f"(got ids {match_ids!r})",
            )

        if len(match_ids) > max_teams_per_slot:
            raise ValueError(
                f"Match {match_num} has too many ids. (got {len(match_ids)}, "
                f"can cope with {max_teams_per_slot})",
            )

        if len(match_ids) % teams_per_game != 0:
            raise ValueError(
                f"Match {match_num} has incompatible number of ids: "
                f"{len(match_ids)} is not a multiple of {teams_per_game}.",
            )

        schedule.append(match_ids)

        for id_ in match_ids:
            if id_ not in ids:
                ids.append(id_)

    return ids, schedule



def max_possible_match_periods(sched_db):
    from datetime import timedelta

    # Compute from the contents of a schedule.yaml the number of league periods
    match_period_length = sched_db['match_slot_lengths']['total']
    total_league_time = sum(
        (
            period['end_time'] - period['start_time']
            for period in sched_db['match_periods']['league']
        ),
        timedelta(),
    )
    return int(total_league_time.total_seconds() // match_period_length)


def command(args):
    import os.path
    import random
    import sys
    from multiprocessing import Pool

    import yaml

    from sr.comp.cli.league_scheduler import Scheduler

    with open(os.path.join(args.compstate, 'arenas.yaml')) as f:
        arenas_db = yaml.load(f)
        arenas = arenas_db['arenas'].keys()
        num_corners = len(arenas_db['corners'])

    with open(os.path.join(args.compstate, 'teams.yaml')) as f:
        teams = yaml.load(f)['teams'].keys()

    with open(os.path.join(args.compstate, 'schedule.yaml')) as f:
        sched_db = yaml.load(f)
        max_periods = max_possible_match_periods(sched_db)

    base_matches = []
    for n in range(args.reschedule_from):
        match_slot = []
        for arena in arenas:
            match_slot.extend(sched_db['matches'][n][arena])
        base_matches.append(match_slot)

    scheduler = Scheduler(
        teams=teams,
        max_match_periods=max_periods,
        arenas=arenas,
        num_corners=num_corners,
        separation=args.spacing,
        max_matchups=args.max_repeated_matchups,
        appearances_per_round=args.appearances_per_round,
        base_matches=base_matches,
        enable_lcg=args.lcg,
    )
    if args.parallel > 1:
        scheduler.lprint(f'Using {args.parallel} threads')
        pool = Pool(args.parallel)

        def get_output(data):
            yaml.dump({'matches': data}, sys.stdout)
            pool.terminate()

        for n in range(args.parallel):
            scheduler.random = random.Random()
            scheduler.tag = f'[Thread {n}] '
            pool.apply_async(scheduler.run, callback=get_output)

        pool.close()
        pool.join()
    else:
        output_data = scheduler.run()
        yaml.dump({'matches': output_data}, sys.stdout)


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'schedule-league',
        help="generate a schedule for a league",
    )
    parser.add_argument(
        'compstate',
        type=str,
        help="competition state git repository",
    )
    parser.add_argument(
        '-s',
        '--spacing',
        type=int,
        default=2,
        help="number of matches between any two appearances by a team",
    )
    parser.add_argument(
        '-r',
        '--max-repeated-matchups',
        type=int,
        default=2,
        help="maximum times any team can face any given other team",
    )
    parser.add_argument(
        '-a',
        '--appearances-per-round',
        type=int,
        default=1,
        help="number of times each team appears in each round",
    )
    parser.add_argument(
        '--lcg',
        action='store_true',
        dest='lcg',
        help="enable LCG permutation",
    )
    parser.add_argument(
        '--parallel',
        type=int,
        default=1,
        help="number of parallel threads",
    )
    parser.add_argument(
        '-f',
        '--reschedule-from',
        type=int,
        default=0,
        help="first match to reschedule from",
    )
    parser.set_defaults(func=command)



DISPLAY_NAME_WIDTH = 18


def first(iterable):
    return next(i for i in iterable)


def command(settings):
    import os.path
    from datetime import datetime, timedelta

    from sr.comp.comp import SRComp

    comp = SRComp(os.path.realpath(settings.compstate))

    num_teams_per_arena = getattr(comp, 'num_teams_per_arena', len(comp.corners))

    matches = comp.schedule.matches
    now = datetime.now(comp.timezone)
    current_matches = list(comp.schedule.matches_at(now))

    if not settings.all:
        time = now - timedelta(minutes=10)

        matches = [
            slot
            for slot in matches
            if first(slot.values()).start_time >= time
        ]

        matches = matches[:int(settings.limit)]

    def teams_str(teams):
        return ":".join(tla.center(5) if tla else "  -  " for tla in teams)

    def print_col(text, last=False):
        print(text, end='|')

    empty_teams = teams_str(" " * num_teams_per_arena)
    teams_len = len(empty_teams)

    print_col(" Num Time  ")
    for a in comp.arenas.values():
        print_col(a.display_name.center(teams_len))
    print_col("Display Name".center(DISPLAY_NAME_WIDTH))
    print()

    arena_ids = comp.arenas.keys()
    for slot in matches:
        m = first(slot.values())
        print_col(f" {m.num:>3} {m.start_time:%H:%M} ")

        for a_id in arena_ids:
            if a_id in slot:
                print_col(teams_str(slot[a_id].teams))
            else:
                print_col(empty_teams)

        print_col(m.display_name.center(DISPLAY_NAME_WIDTH))

        if m in current_matches:
            print(" *")
        else:
            print()


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'show-schedule',
        help="show the match schedule",
    )
    parser.add_argument(
        'compstate',
        help="competition state repo",
    )
    parser.add_argument(
        '--all',
        action='store_true',
        help="show all matches, not just the upcoming ones (ignores --limit)",
    )
    parser.add_argument(
        '--limit',
        default=15,
        help="how many matches to show (default: %(default)s)",
    )
    parser.set_defaults(func=command)



from contextlib import contextmanager
from typing import Any, Iterator, Optional, Sequence, Type

API_TIMEOUT_SECONDS = 3
SSH_TIMEOUT_SECONDS = 2
DEPLOY_USER = 'srcomp'
BOLD = '\033[1m'
FAIL = '\033[91m'
OKBLUE = '\033[94m'
ENDC = '\033[0m'


def ssh_connection(host):
    from paramiko import AutoAddPolicy, SSHClient

    client = SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(AutoAddPolicy())
    client.connect(host, username=DEPLOY_USER, timeout=SSH_TIMEOUT_SECONDS)
    return client


def format_fail(*args: object) -> str:
    msg = " ".join(map(str, args))
    return BOLD + FAIL + msg + ENDC


@contextmanager
def exit_on_exception(
    msg: str = '{0}',
    kind: Type[Exception] = Exception,
) -> Iterator[None]:
    try:
        yield
    except kind as e:
        print_fail(msg.format(e))
        exit(1)


def print_fail(*args: object, **kargs: Any) -> None:
    print(format_fail(*args), **kargs)


def print_buffer(buf):
    prefix = '> '
    print(prefix + prefix.join(buf.readlines()).strip())


def get_input(prompt: str) -> str:
    # Wrapper to simplify mocking
    return input(prompt)


def query(
    question: str,
    options: Sequence[str],
    default: Optional[str] = None,
) -> str:
    if default:
        assert default in options

    options = [o.upper() if o == default else o.lower() for o in options]
    assert len(set(options)) == len(options)
    opts = '/'.join(options)

    query = format_fail(f"{question.rstrip()} [{opts}]: ")

    while True:
        # Loop until we get a suitable response from the user
        resp = get_input(query).lower()

        if resp in options:
            return resp

        # If there's a default value, use that
        if default:
            return default


def query_bool(question: str, default_val: Optional[bool] = None) -> bool:
    options = ('y', 'n')
    if default_val is True:
        default = 'y'  # type: Optional[str]
    elif default_val is False:
        default = 'n'
    else:
        default = None
    return query(question, options, default) == 'y'


def query_warn(msg: object) -> None:
    if not query_bool(f"Warning: {msg}. Continue?", False):
        exit(1)


def ref_compstate(host: str) -> str:
    return f'ssh://{DEPLOY_USER}@{host}/~/compstate.git'


def deploy_to(compstate, host, revision, verbose):
    print(BOLD + f"Deploying to {host}:" + ENDC)

    # Make connection early to check if host is up.
    with ssh_connection(host) as client:
        # Push the repo
        url = ref_compstate(host)
        # Make a new branch for this revision so that it's visible to
        # anything which fetches the repo; use the revision id in the
        # branch name to avoid race conditions without needing to come
        # up with our own unique identifier.
        # This also means we don't need to worry about whether or not the
        # revision exists in the target, since this push will simply no-op
        # if it's already present
        revspec = '{0}:refs/heads/deploy-{0}'.format(revision)
        with exit_on_exception(kind=RuntimeError):
            compstate.push(
                url,
                revspec,
                err_msg=f"Failed to push to {host}.",
            )

        cmd = f"./update '{revision}'"
        _, stdout, stderr = client.exec_command(cmd)
        retcode = stdout.channel.recv_exit_status()

        if verbose or retcode != 0:
            print_buffer(stdout)

        print_buffer(stderr)

        return retcode


def get_deployments(compstate):
    with exit_on_exception("Failed to get deployments from state ({0})."):
        return compstate.deployments


def get_current_state(host):
    from urllib.request import urlopen

    import simplejson as json

    url = f'http://{host}/comp-api/state'
    try:
        page = urlopen(url, timeout=API_TIMEOUT_SECONDS)
        raw_state = json.load(page)
    except Exception as e:
        print(e)
        return None
    else:
        return raw_state['state']


def check_host_state(compstate, host, revision, verbose):
    """
    Compares the host state to the revision we want to deploy. If the
    host's state isn't in the history of the deploy revision then various
    options are presented to the user.

    Returrns whether or not to skip deploying to the host.
    """
    SKIP = True
    UPDATE = False
    if verbose:
        print(f"Checking host state for {host} (timeout {API_TIMEOUT_SECONDS} seconds).")
    state = get_current_state(host)
    if not state:
        if query_bool(
            f"Failed to get state for {host}, cannot advise about history. Deploy anyway?",
            True,
        ):
            return UPDATE
        else:
            return SKIP

    if state == revision:
        print(f"Host {host} already has requested revision ({revision[:8]})")
        return SKIP

    # Ideal case:
    if compstate.has_ancestor(state):
        return UPDATE

    # Check for unknown commit
    if not compstate.has_commit(state):
        if query_bool(f"Host {host} has unknown state '{state}'. Try to fetch it?", True):
            compstate.fetch('origin', quiet=True)
            compstate.fetch(ref_compstate(host), quiet=True)

    # Old revision:
    if compstate.has_descendant(state):
        if query_bool(f"Host {host} has more recent state '{state}'. Deploy anyway?", True):
            return UPDATE
        else:
            return SKIP

    # Some other revision:
    if compstate.has_commit(state):
        if query_bool(f"Host {host} has sibling state '{state}'. Deploy anyway?", True):
            return UPDATE
        else:
            return SKIP

    # An unknown state
    if query_bool(f"Host {host} has unknown state '{state}'. Deploy anyway?", True):
        return UPDATE
    else:
        return SKIP


def require_no_changes(compstate):
    if compstate.has_changes:
        print_fail(
            "Cannot deploy state with local changes.",
            "Commit or remove them and re-run.",
        )
        compstate.show_changes()
        exit(1)


def require_valid(compstate):
    from sr.comp.validation import validate

    with exit_on_exception("State cannot be loaded: {0}"):
        comp = compstate.load()

    num_errors = validate(comp)
    if num_errors:
        query_warn("State has validation errors (see above)")


def run_deployments(args, compstate, hosts):
    revision = compstate.rev_parse('HEAD')
    for host in hosts:
        if not args.skip_host_check:
            skip_host = check_host_state(compstate, host, revision, args.verbose)
            if skip_host:
                print(BOLD + f"Skipping {host}." + ENDC)
                continue

        retcode = deploy_to(compstate, host, revision, args.verbose)
        if retcode != 0:
            # TODO: work out if it makes sense to try to rollback here?
            print_fail(f"Failed to deploy to '{host}' (exit status: {retcode}).")
            exit(retcode)

    print(BOLD + OKBLUE + "Done" + ENDC)


def command(args):
    from sr.comp.raw_compstate import RawCompstate

    compstate = RawCompstate(args.compstate, local_only=False)
    hosts = get_deployments(compstate)

    require_no_changes(compstate)
    require_valid(compstate)

    run_deployments(args, compstate, hosts)


def add_options(parser):
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument(
        '--skip-host-check',
        action='store_true',
        help="skips checking the current state of the hosts",
    )


def add_subparser(subparsers):
    help_msg = "Deploy a given competition state to all known hosts"
    parser = subparsers.add_parser(
        'deploy',
        help=help_msg,
        description=help_msg,
    )
    add_options(parser)
    parser.add_argument('compstate', help="competition state repository")
    parser.set_defaults(func=command)



from sr.comp.cli import add_delay, deploy


def command(args):
    from sr.comp.raw_compstate import RawCompstate

    compstate = RawCompstate(args.compstate, local_only=False)
    hosts = deploy.get_deployments(compstate)

    deploy.require_no_changes(compstate)

    if not args.no_pull:
        with deploy.exit_on_exception():
            compstate.pull_fast_forward()

    how_long, when = add_delay.command(args)

    if args.when != 'now':
        msg = f"Confirm adding {how_long} delay at {when}"
        if not deploy.query_bool(msg, default_val=True):
            print("Leaving state with local modifications")
            exit()

    deploy.require_valid(compstate)

    with deploy.exit_on_exception(kind=RuntimeError):
        compstate.stage('schedule.yaml')
        msg = f"Adding {args.how_long} delay at {when}"
        compstate.commit(msg)

    deploy.run_deployments(args, compstate, hosts)


def add_subparser(subparsers):
    help_msg = "Add and deploy a delay to the competition"
    parser = subparsers.add_parser('delay', help=help_msg, description=help_msg)
    parser.add_argument(
        '--no-pull',
        action='store_true',
        help="skips updating to the latest revision",
    )
    deploy.add_options(parser)
    parser.add_argument('compstate', help="competition state repository")
    add_delay.add_arguments(parser)
    parser.set_defaults(func=command)



"""
Import a league.yaml from a schedule file.

A schedule file specifies matches one-per-line, as follows:

A 'match' consists of a number of unique identifiers separated by pipe
characters. The total number of identifiers in the file should be equal
to or greater than the number of teams in the compstate.

The number of identifiers in a given match must be a multiple of the
number of teams per game (currently 4), up to the number of arenas in
the compstate.

Whitespace (other than newlines) within the file is ignored, as is any
content to the right of a hash character (#), including the hash. As a
result hash characters may be used to start line comments.

Example schedules for 48, 52 or 56 teams are available  at:
https://github.com/PeterJCLaw/srobo-schedules/tree/master/seed_schedules
"""

import argparse
from pathlib import Path
from typing import Dict, Iterable

from sr.comp.cli.import_schedule import loading, teams_mapping
from sr.comp.cli.import_schedule.types import Configuration, RawMatch
from sr.comp.types import MatchNumber


def get_first_match_number(existing_match_numbers: Iterable[MatchNumber]) -> MatchNumber:
    if not existing_match_numbers:
        return MatchNumber(0)

    return MatchNumber(max(existing_match_numbers) + 1)


def get_configuration(
    compensate_path: Path,
    team_order_strategy: teams_mapping.Strategy,
    existing_match_numbers: Iterable[MatchNumber],
) -> Configuration:
    # Grab the teams and arenas
    try:
        team_ids, arena_ids, teams_per_game = loading.load_teams_areans(compensate_path)
    except Exception as e:
        print(f"Failed to load existing state ({e}).")
        print("Make it valid (consider removing the league.yaml and layout.yaml)")
        print("and try again.")
        exit(1)

    # Semi-randomise
    team_ids = teams_mapping.order_teams(
        compensate_path,
        team_ids,
        team_order_strategy,
    )

    return Configuration(
        arena_ids,
        team_ids,
        teams_per_game,
        get_first_match_number(existing_match_numbers),
    )


def command(args: argparse.Namespace) -> None:
    from sr.comp.cli.import_schedule import core

    with open(args.schedule) as sfp:
        schedule_lines = loading.tidy(sfp.readlines())

    league_yaml = loading.league_yaml_path(args.compstate)
    existing_matches: Dict[MatchNumber, RawMatch] = {}
    if args.extend:
        existing_matches = loading.load_league_yaml(league_yaml)

    config = get_configuration(
        args.compstate,
        args.team_order_strategy,
        existing_matches.keys(),
    )

    matches, bad_matches = core.build_schedule(
        config,
        schedule_lines,
        args.ignore_ids,
    )

    # Print any warnings about the matches
    for bad_match in bad_matches:
        print(
            f"Warning: match {bad_match.arena}:{bad_match.num} only has "
            f"{bad_match.num_teams} teams.",
        )

    # Save the matches to the file
    loading.dump_league_yaml({**existing_matches, **matches}, league_yaml)


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    parser = subparsers.add_parser(
        'import-schedule',
        help="Import a league.yaml file from a schedule file",
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '-i',
        '--ignore-ids',
        type=loading.parse_ids,
        help="comma separated list of ids to ignore",
    )
    parser.add_argument(
        '--extend',
        default=False,
        action='store_true',
        help=(
            "Whether to replace (the defualt) or extend the existing league "
            "with the matches in the given schedule file."
        ),
    )
    parser.add_argument(
        '--team-order-strategy',
        choices=teams_mapping.Strategy,
        default=teams_mapping.Strategy.AUTO,
        type=teams_mapping.Strategy,
        help="How to map schedule ids to TLAs",
    )
    parser.add_argument('compstate', type=Path, help="competition state repository")
    parser.add_argument('schedule', type=Path, help="schedule to import")
    parser.set_defaults(func=command)



import argparse
from collections import defaultdict


class ScheduleGenerator:
    def __init__(self, target, arenas, state):
        from reportlab.pdfgen import canvas  # type: ignore[import]

        self.canvas = canvas.Canvas(target)
        self.state = state
        self.width = 595
        self.height = 842
        self.margin = 50
        self.page_number = 0
        self.arenas = arenas
        self.columns = 2 + 4 * len(arenas)

    def start_page(self, title="Match Schedule"):
        self.row_height = 800
        if self.page_number != 0:
            self.canvas.showPage()
        self.page_number += 1

        self.draw_header(title)
        self.draw_footer()
        self.draw_vertical_bars()
        self.draw_column_headings()

    def draw_header(self, text):
        self.canvas.setFont('Helvetica', 10)
        self.canvas.drawCentredString(self.width * 0.5, 820, text)

    def draw_footer(self):
        self.canvas.setFont('Helvetica', 8)
        self.canvas.drawCentredString(
            self.width * 0.5,
            10,
            "Page {} • Generated from state {}".format(
                self.page_number,
                self.state[:7],
            ),
        )

    def draw_vertical_bars(self):
        if len(self.arenas) == 1:
            cols = (200,)
        elif len(self.arenas) == 2:
            cols = (134, 353)
        else:
            raise RuntimeError(
                "Unexpected number of arenas: {}".format(len(self.arenas)),
            )

        for x in cols:
            self.canvas.line(x, 30, x, 810)

    def draw_column_headings(self):
        headings = [("Number", 'white', True), ("Time", 'white', True)]
        for arena in self.arenas.values():
            headings += [
                (arena.display_name, 'white', True),
                "",
                "",
                "",
            ]
        self.add_line(headings)

    def add_line(self, line):
        if len(line) != self.columns:
            raise ValueError("Incorrect column count")
        for i, cell in enumerate(line):
            if isinstance(cell, tuple):
                text = cell[0]
                background = cell[1]
                try:
                    bold = cell[2]
                except IndexError:
                    bold = False
            else:
                text = cell
                background = None
                bold = False

            if bold:
                self.canvas.setFont('Helvetica-Bold', 11)
            else:
                self.canvas.setFont('Helvetica', 10)

            centre_x = self.margin + i * (self.width - 2 * self.margin) / (self.columns - 1)
            centre_y = self.row_height

            if background is not None:
                self.canvas.setFillColor(background)
                self.canvas.rect(
                    centre_x - 20,
                    centre_y - 4,
                    40,
                    14,
                    stroke=False,
                    fill=True,
                )
                self.canvas.setFillColor('black')

            self.canvas.drawCentredString(centre_x, centre_y, text)
        self.canvas.line(
            self.margin * 0.5,
            self.row_height - 3.5,
            self.width - (self.margin * 0.5),
            self.row_height - 3.5,
        )
        self.row_height -= 14

    @staticmethod
    def _get_periods(competition, numbers=None):
        comp_periods = competition.schedule.match_periods
        if numbers is None:
            return comp_periods

        periods = []
        for n in numbers:
            periods.append(comp_periods[n])

        return periods

    @staticmethod
    def _get_shepherds(raw_compstate, numbers=None, combined=False):
        def split(shepherds):
            if combined:
                return [shepherds]
            return [(x,) for x in shepherds]

        comp_shepherds = raw_compstate.load_shepherds()
        if numbers is None:
            return split(comp_shepherds)

        shepherds = []
        for n in numbers:
            shepherds.append(comp_shepherds[n])
        return split(shepherds)

    @staticmethod
    def _get_locations(raw_compstate, names=None):
        comp_locations = raw_compstate.layout['teams']
        if names is None:
            return comp_locations

        locations = []
        for name in names:
            for location in comp_locations:
                if location['name'] == name:
                    locations.append(location)
                    break
            else:
                raise KeyError(name)

        return locations

    @staticmethod
    def _get_page_title(period, shepherds, locations, include_locations=True):
        title = str(period)

        if shepherds:
            title += " • Shepherd {}".format(
                ", ".join(
                    shepherd.get('name', "#{}".format(i + 1))
                    for i, shepherd in enumerate(shepherds)
                ),
            )

        if include_locations and locations:
            title += " • {}".format(
                ", ".join(x['display_name'] for x in locations),
            )

        return title

    @staticmethod
    def _match_suitable_for_locations(slot, locations):
        suitable_teams = ['???']
        for location in locations:
            suitable_teams += location['teams']

        for match in slot.values():
            for team in match.teams:
                if team in suitable_teams:
                    return True
        return False

    def _generate(self, period, shepherds, locations, include_locations_in_title, is_plain):
        def find_shepherd_number(team):
            if shepherds is None:
                return None
            for i, shepherd in enumerate(shepherds):
                if team in shepherd['teams']:
                    return i
            return None

        team_colours = {}
        if shepherds:
            for shepherd in shepherds:
                for team in shepherd['teams']:
                    team_colours[team] = shepherd['colour']

        def get_team_cell(team):
            if team == '???':
                team = '____'
            elif not team:
                team = '–'

            if is_plain:
                return team

            colour = team_colours.get(team, 'white')

            # If the shepherd for this team needs to get at least 4
            # teams during this slot them embolden all their teams
            bold = shepherd_counts.get(find_shepherd_number(team), 0) >= 4

            cell = (team, colour, bold)
            return cell

        title = self._get_page_title(
            period,
            shepherds,
            locations,
            include_locations_in_title,
        )
        self.start_page(title)

        n = 0
        for slot in period.matches:
            # Build map of shepherd id -> total teams they need to fetch this slot
            shepherd_counts = defaultdict(int)
            for match in slot.values():
                for team in match.teams:
                    num = find_shepherd_number(team)
                    if num is not None:
                        shepherd_counts[num] += 1

            if self._match_suitable_for_locations(slot, locations):
                cells = ['', '']
                for arena in self.arenas:
                    match = slot.get(arena)
                    if match is not None:
                        cells += list(map(get_team_cell, match.teams))
                        cells[0] = str(match.num)
                        cells[1] = str(match.start_time.strftime('%H:%M'))
                    else:
                        cells += ['–', '–', '–', '–']
                self.add_line(cells)
                n += 1

            if n % 65 == 65:
                self.start_page(title)

    def generate(
        self,
        competition,
        raw_compstate,
        period_numbers,
        shepherd_numbers,
        location_names,
        is_plain,
        combine_shepherds,
    ):
        periods = self._get_periods(competition, period_numbers)
        shepherd_groups = self._get_shepherds(
            raw_compstate, shepherd_numbers,
            combine_shepherds,
        )
        locations = self._get_locations(raw_compstate, location_names)

        for shepherds in shepherd_groups:
            for period in periods:
                self._generate(
                    period, shepherds, locations,
                    location_names is not None, is_plain,
                )

    def write(self):
        self.canvas.save()


def command(settings):
    import os.path

    from sr.comp.comp import SRComp
    from sr.comp.raw_compstate import RawCompstate

    comp = SRComp(os.path.realpath(settings.compstate))
    raw_comp = RawCompstate(
        os.path.realpath(settings.compstate),
        local_only=True,
    )

    generator = ScheduleGenerator(
        settings.output,
        arenas=comp.arenas,
        state=comp.state,
    )

    generator.generate(
        comp,
        raw_comp,
        settings.periods,
        settings.shepherds,
        settings.locations,
        settings.plain,
        settings.shepherds_combined,
    )

    generator.write()


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'print-schedule',
        help="print a shepherding sheet",
    )
    parser.add_argument('compstate', help="competition state repository")
    parser.add_argument(
        '-o',
        '--output',
        help="output file",
        type=argparse.FileType('wb'),
        required=True,
    )
    parser.add_argument(
        '--plain',
        action='store_true',
        help="output the schedule without any colouring or emboldening",
    )
    parser.add_argument(
        '-p',
        '--periods',
        type=int,
        nargs='+',
        help="specify periods by number",
    )
    parser.add_argument(
        '-s',
        '--shepherds',
        type=int,
        nargs='+',
        help="specify shepherds by number",
    )
    parser.add_argument(
        '-c',
        '--shepherds-combined',
        action='store_true',
        default=False,
        help=(
            "combine the highlighting of shepherds onto a single sheet (default "
            "is to print a separate sheet for each shepherd)"
        ),
    )
    parser.add_argument(
        '-l',
        '--locations',
        nargs='+',
        help="specify locations by name",
    )
    parser.set_defaults(func=command)



"""
Run the SRComp Scorer UI. Requires sr.comp.scorer to be installed.
"""


def find_unused_port():
    import socket

    sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    sock.bind(('::1', 0))
    print(sock.getsockname())
    addr, port, flowinfo, scopeid = sock.getsockname()
    sock.close()
    return port


def command(settings):
    import threading
    import time
    import webbrowser

    try:
        import sr.comp.scorer  # type: ignore[import]
    except ImportError:
        print("sr.comp.scorer not installed.")
        exit(1)

    port = find_unused_port()
    app = sr.comp.scorer.app
    app.config['COMPSTATE'] = settings.compstate
    app.config['COMPSTATE_LOCAL'] = not settings.push_changes

    def browse():
        time.sleep(1.5)
        webbrowser.open(f'http://localhost:{port}/')

    thread = threading.Thread(target=browse)
    thread.start()

    try:
        app.run(
            host='::1',
            port=port,
            debug=False,
            passthrough_errors=True,
        )
    except KeyboardInterrupt:
        pass


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'score',
        help=__doc__.strip().splitlines()[0],
        description=__doc__,
    )
    parser.add_argument('compstate', help="competition state repository")
    parser.add_argument(
        '-p',
        '--push-changes',
        action='store_true',
        help="send commits upstream to origin/master",
    )
    parser.set_defaults(func=command)



__description__ = "List available MIDI output ports."


def command(args):
    import mido  # type: ignore[import]

    ports = mido.get_output_names()
    print(len(ports), "outputs:")
    for port in ports:
        print("-", port)


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'list-midi-ports', help=__description__,
        description=__description__,
    )
    parser.set_defaults(func=command)



"""Update the layout based on a list of teams.
The user is responsible for ensuring that the ordering and groupings are
correct in the layout _before_ running this. Groups of teams in the
original are simply replaced with equivalently sized groups.
Any excess teams in the list are added to the final group."""


class Takeable:
    def __init__(self, source):
        self._source = source
        self._idx = 0

    @property
    def has_more(self):
        return self._idx < len(self._source)

    @property
    def remainder(self):
        return self._source[self._idx:]

    def take(self, n):
        start = self._idx
        self._idx = end = start + n
        return self._source[start:end]


def command(settings):
    import os.path

    from sr.comp.cli import yaml_round_trip as rtyaml

    layout_yaml = os.path.join(settings.compstate, 'layout.yaml')
    layout = rtyaml.load(layout_yaml)
    layout_teams = layout['teams']

    with open(settings.teams_list) as tlf:
        teams_list = []
        for line in tlf.readlines():
            tla = line.split('#', 1)[0].strip()
            if tla:
                teams_list.append(tla)

    teams = Takeable(teams_list)

    for place in layout_teams:
        # Ensure we replace the content of the list, but not the list
        # itself so that the file's own layout is preserved
        loc_teams = place['teams']
        loc_teams[:] = teams.take(len(loc_teams))

    if teams.has_more:
        layout_teams[-1]['teams'] += teams.remainder

    rtyaml.dump(layout_yaml, layout)

    print("Layout updated. You should consider re-importing the schedule now.")


def add_subparser(subparsers):
    parser = subparsers.add_parser('update-layout', help=__doc__)
    parser.add_argument('compstate', help="competition state repository")
    parser.add_argument('teams_list', help="file containing the list of teams")
    parser.set_defaults(func=command)



__description__ = "Summaries the teams scoring the most match points"


def command(settings):
    import os.path
    from collections import defaultdict
    from functools import partial
    from itertools import chain

    from sr.comp.comp import SRComp

    comp = SRComp(os.path.realpath(settings.compstate))

    all_scores = (comp.scores.tiebreaker, comp.scores.knockout, comp.scores.league)
    all_points = dict(chain.from_iterable(s.game_points.items() for s in all_scores))

    points_map = defaultdict(partial(defaultdict, list))

    for match, points in all_points.items():
        for tla, team_points in points.items():
            points_map[team_points][tla].append(match)

    count = len(points_map)

    for idx, (points, team_info) in enumerate(sorted(points_map.items())):
        if idx + 2 < count:
            print(f"{len(team_info):>3} teams scored {points}")
        else:
            print()
            print(f"The following {len(team_info)} team(s) scored {points} points:")
            for tla, matches in team_info.items():
                print(f"- {tla} in match(es): " + ", ".join(
                    "{}{}".format(*x)
                    for x in matches
                ))


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'top-match-points',
        help=__description__,
        description=__description__,
    )
    parser.add_argument(
        'compstate',
        help="competition state repo",
    )
    parser.set_defaults(func=command)



"""
Shift the matches to start at the current time.

This is primarily aimed for development of SRComp related tooling, as a way to
reset the matches in a compstate to start at a convenient time.
"""


def command(args):
    from datetime import datetime, timedelta

    from .yaml_round_trip import dump as dump_yaml, load as load_yaml

    schedule = load_yaml(str(args.compstate / 'schedule.yaml'))

    old_start = schedule['match_periods'][args.focus][0]['start_time']
    new_start = datetime.now(old_start.tzinfo)
    # round to 1-2 minutes ahead
    new_start -= timedelta(
        seconds=new_start.second,
        microseconds=new_start.microsecond,
    )
    new_start += timedelta(minutes=2)

    dt = new_start - old_start

    for group in schedule['match_periods'].values():
        for entry in group:
            entry['start_time'] += dt
            entry['end_time'] += dt
            if 'max_end_time' in entry:
                entry['max_end_time'] += dt

    dump_yaml(str(args.compstate / 'schedule.yaml'), schedule)

    with (args.compstate / '.update-pls').open('w'):
        pass
    print(f"Shifted matches by {dt}")


def add_subparser(subparsers):
    from pathlib import Path

    parser = subparsers.add_parser(
        'shift-matches',
        help=__doc__.strip().splitlines()[0],
        description=__doc__,
    )
    parser.add_argument(
        'compstate',
        type=Path,
        help="competition state repository",
    )
    parser.add_argument(
        'focus',
        choices=('league', 'knockout'),
        help="match period to focus",
    )
    parser.set_defaults(func=command)



"""
Run a command for each of the given matches.

The matches can be specified as a comma-separated list and/or dash-separated
ranges. The commands will be run in the numeric order of matches (rather than
the given order) and only once per match.

Where a compstate has multiple arenas the command will be run for each arena, in
the order the arenas are defined within the compstate. Alternatively a specific
arena can be specified to limit the commands to matches in that arena.

A number of placeholders are available for use in your command, so you can
control how information about each match is passed to the command. The TLAS
placeholder will always include enough entries for every zone in the arena,
using '-' to represent an empty zone.
"""

import argparse
import sys
from typing import Callable, Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    from sr.comp.match_period import Match


def get_tlas(match: 'Match') -> List[str]:
    return [
        x if x is not None else '-'
        for x in match.teams
    ]


PLACEHOLDERS = {
    'ARENA': lambda x: [x.arena],
    'NUMBER': lambda x: [str(x.num)],
    'TLAS': get_tlas,
    'TYPE': lambda x: [str(x.type.value)],
}  # type: Dict[str, Callable[[Match], List[str]]]


class PlaceholderExpander:
    def __init__(self, match: 'Match') -> None:
        self.match = match

    @staticmethod
    def validate(value: str) -> None:
        if value.startswith('@') and value[1:] not in PLACEHOLDERS:
            print(f"Warning: unrecognised value {value!r}.", file=sys.stderr)

    def __getitem__(self, key: str) -> str:
        return ' '.join(PLACEHOLDERS[key](self.match))

    def expand(self, value: str) -> List[str]:
        if value.startswith('@'):
            key = value[1:]
            fn = PLACEHOLDERS.get(key)
            if fn:
                return fn(self.match)

        return [value.format_map(self)]


def replace_placeholders(match: 'Match', command: List[str]) -> List[str]:
    import itertools

    expander = PlaceholderExpander(match)
    return list(itertools.chain.from_iterable(
        expander.expand(x)
        for x in command
    ))


def command(args):
    import subprocess

    from sr.comp.comp import SRComp

    from .deploy import print_fail

    compstate = SRComp(args.compstate)

    if args.arena:
        if args.arena not in compstate.arenas:
            print("{} is not a valid arena, choose from {}".format(
                args.arena,
                ", ".join(compstate.arenas.keys()),
            ))

    for part in args.command:
        PlaceholderExpander.validate(part)

    try:
        for match_number in sorted(args.matches):
            for arena, match in compstate.schedule.matches[match_number].items():
                if args.arena not in (arena, None):
                    continue

                command = replace_placeholders(match, args.command)
                subprocess.check_call(command)

    except subprocess.CalledProcessError as e:
        print_fail(str(e))
        exit(1)


def add_options(parser):
    from sr.comp.matches import parse_ranges

    parser.add_argument(
        '--arena',
        default=None,
        help=(
            "Limit to just one arena. By default the command is run for each "
            "arena in turn."
        ),
    )
    parser.add_argument(
        'matches',
        type=parse_ranges,
        help="List of matches or match ranges, for example '1,3-5'.",
    )
    parser.add_argument(
        'command',
        nargs='+',
        help=(
            "Command to run. Supports the following placeholders: {}. "
            "Placeholders spelled like {{THIS}} will be replaced as strings "
            "anywhere within the command arguments. Placeholders which are an "
            "argument on their own and spelled exactly as @THIS will expand to "
            "one or more replacement arguments when the command is run.".format(
                ", ".join(PLACEHOLDERS.keys()),
            )
        ),
    )


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'for-each-match',
        help="Run a command for each of the given matches.",
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('compstate', help="competition state repository")
    add_options(parser)
    parser.set_defaults(func=command)



def first(iterable):
    return next(i for i in iterable)


def counter_to_string(cntr):
    string = ", ".join("{1} {0}".format(*item) for item in cntr.items())
    return string


def command(args):
    from collections import Counter

    from sr.comp.comp import SRComp

    comp = SRComp(args.compstate)

    print("Number of arenas: {} ({})".format(
        len(comp.arenas),
        ", ".join(comp.arenas.keys()),
    ))

    print("Number of teams: {} ({} rookies)".format(
        len(comp.teams),
        sum(1 for t in comp.teams.values() if t.rookie),
    ))

    slots_by_type = Counter(
        first(slot.values()).type.value
        for slot in comp.schedule.matches
    )
    slots_by_type_str = counter_to_string(slots_by_type)

    assert sum(slots_by_type.values()) == len(comp.schedule.matches)

    print("Number of match slots: {} ({})".format(
        len(comp.schedule.matches),
        slots_by_type_str,
    ))

    games_by_type = Counter(
        game.type.value
        for slot in comp.schedule.matches
        for game in slot.values()
    )
    games_by_type_str = counter_to_string(games_by_type)

    print("Number of games: {} ({})".format(
        sum(games_by_type.values()),
        games_by_type_str,
    ))


def add_subparser(subparsers):
    help_msg = "Show summary data about a compstate"

    parser = subparsers.add_parser('summary', help=help_msg, description=help_msg)
    parser.add_argument('compstate', help="competition state repository")
    parser.set_defaults(func=command)



import collections
from typing import (
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from sr.comp.types import ArenaName, MatchNumber, TLA

from . import loading
from .types import BadMatch, Configuration, ID, RawMatch

T = TypeVar('T')


def chunks_of_size(list_: List[T], size: int) -> Iterator[List[T]]:
    if len(list_) % size != 0:
        raise ValueError(
            "Unable to chunk list whose size is not evenly divisble by given "
            f"size {size}.",
        )

    list_ = list_[:]
    while len(list_):
        chunk = []
        for _ in range(size):
            chunk.append(list_.pop(0))
        yield chunk


def ignore_ids(ids: List[ID], ids_to_remove: List[ID]) -> None:
    for i in ids_to_remove:
        ids.remove(i)


def get_id_subsets(ids: Collection[T], limit: int) -> Iterator[Collection[T]]:
    num_ids = len(ids)

    extra = num_ids - limit

    if extra == 0:
        # Only one posibility -- use all of them
        yield ids

    elif extra == 1:
        for idx in range(len(ids)):
            ids_clone = list(ids)
            ids_clone.pop(idx)
            yield ids_clone

    elif extra == 2:
        for idx1 in range(len(ids)):
            for idx2 in range(idx1 + 1, len(ids)):
                ids_clone = list(ids)
                ids_clone.pop(idx2)
                ids_clone.pop(idx1)
                yield ids_clone

    elif extra == 3:
        for idx1 in range(len(ids)):
            for idx2 in range(idx1 + 1, len(ids)):
                for idx3 in range(idx2 + 1, len(ids)):
                    ids_clone = list(ids)
                    ids_clone.pop(idx3)
                    ids_clone.pop(idx2)
                    ids_clone.pop(idx1)
                    yield ids_clone

    else:
        # TODO: consider generalising the above or adding more handling
        raise Exception(f"Too many empty slots to compensate for ({extra}).")


def build_id_team_maps(ids: List[ID], team_ids: Sequence[TLA]) -> Iterator[Dict[ID, TLA]]:
    # If there are more ids than team_ids we want to ensure that we minimize
    # the number of matches which have empty places and also the number of
    # empty places in any given match.
    # This function generates possible mappings of ids to teams as needed
    # in order to explore excluding different ids.
    #
    # Note: this function does _not_ explore mapping the same subset of
    # ids to the given teams since that doesn't achieve any changes in
    # which matches have empty spaces.

    for id_subset in get_id_subsets(ids, len(team_ids)):
        yield dict(zip(id_subset, team_ids))


def build_matches(
    id_team_map: Dict[ID, TLA],
    schedule: List[List[ID]],
    arena_ids: Collection[ArenaName],
    teams_per_game: int,
    first_match_number: int,
) -> Tuple[
    Dict[MatchNumber, RawMatch],
    List[BadMatch],
]:
    matches = {}
    bad_matches = []
    for match_num, match_ids in enumerate(schedule, start=first_match_number):
        match_teams = [id_team_map.get(id_) for id_ in match_ids]
        games = chunks_of_size(match_teams, teams_per_game)

        matches[MatchNumber(match_num)] = match = dict(zip(arena_ids, games))

        # Check that the match has enough actual teams; warn if not
        for arena, teams in match.items():
            num_teams = len(set(teams) - set([None]))
            if num_teams <= (teams_per_game / 2):
                bad_matches.append(BadMatch(arena, MatchNumber(match_num), num_teams))

    return matches, bad_matches


def are_better_matches(
    best: List[BadMatch],
    new: List[BadMatch],
    teams_per_game: int,
) -> bool:
    def get_empty_places_map(bad_matches: List[BadMatch]) -> Mapping[int, int]:
        empty_places_map: Dict[int, int] = collections.Counter()
        for bad_match in bad_matches:
            num_empty = teams_per_game - bad_match.num_teams
            empty_places_map[num_empty] += 1
        return empty_places_map

    best_map = get_empty_places_map(best)
    new_map = get_empty_places_map(new)

    possible_empty_places = set(list(best_map.keys()) + list(new_map.keys()))

    # Even single matches with lots of empty slots are bad
    for num_empty_places in sorted(possible_empty_places, reverse=True):
        if new_map[num_empty_places] < best_map[num_empty_places]:
            return True

    return False


def get_best_fit(
    config: Configuration,
    ids: List[ID],
    schedule: List[List[ID]],
) -> Tuple[
    Dict[MatchNumber, RawMatch],
    List[BadMatch],
]:
    best: Optional[Tuple[
        Dict[MatchNumber, RawMatch],
        List[BadMatch],
    ]] = None
    for id_team_map in build_id_team_maps(ids, config.team_ids):
        matches, bad_matches = build_matches(
            id_team_map,
            schedule,
            config.arena_ids,
            config.teams_per_game,
            config.first_match_number,
        )

        if len(bad_matches) == 0:
            # Nothing bad about these, ship them
            return matches, bad_matches

        if best is None or are_better_matches(best[1], bad_matches, config.teams_per_game):
            best = (matches, bad_matches)

    assert best is not None

    return best


def build_schedule(
    config: Configuration,
    schedule_lines: List[str],
    ids_to_ignore: List[ID],
) -> Tuple[
    Dict[MatchNumber, RawMatch],
    List[BadMatch],
]:
    # Collect up the ids used
    ids, schedule = loading.load_ids_schedule(
        schedule_lines,
        num_arenas=config.num_arenas,
        teams_per_game=config.teams_per_game,
    )

    # Ignore any ids we've been told to
    if ids_to_ignore:
        ignore_ids(ids, ids_to_ignore)

    # Sanity checks
    if len(ids) < config.num_teams:
        raise ValueError(
            f"Not enough places in the schedule (need {len(ids)}, got {config.num_teams}).",
        )

    # Get matches
    matches, bad_matches = get_best_fit(config, ids, schedule)

    return matches, bad_matches



import collections
import enum
from pathlib import Path
from typing import List

from sr.comp.types import TLA


class Strategy(enum.Enum):
    AUTO = 'auto'
    LAYOUT = 'layout'
    RANDOM = 'random'

    def __str__(self) -> str:
        return self.value


def order_teams_randomly(team_ids: List[TLA]) -> List[TLA]:
    """
    Order teams using our stable random logic.
    """
    from sr.comp.knockout_scheduler.stable_random import Random

    random = Random()
    random.seed("".join(team_ids).encode())
    random.shuffle(team_ids)
    return team_ids


def order_teams_by_location(layout_yaml: Path, team_ids: List[TLA]) -> List[TLA]:
    """
    Order teams by location, such that the order of appearances in the matches
    is equialent to the ordering in the layout.

    This is useful as it should mean that the scrutineers can move around the
    venue easily visiting teams in the order which they first appear in matches.
    """
    import yaml

    from sr.comp.validation import join_and

    with open(layout_yaml) as lf:
        layout_raw = yaml.load(lf)
        layout = layout_raw['teams']

    ordered_teams = []
    for group in layout:
        ordered_teams += group['teams']

    layout_teams = set(ordered_teams)

    if len(layout_teams) != len(ordered_teams):
        duplicates = [x for x, y in collections.Counter(ordered_teams).items() if y > 1]
        raise ValueError(f"Some teams appear twice in the layout! {join_and(duplicates)}")

    all_teams = set(team_ids)
    missing = all_teams - layout_teams
    if missing:
        raise ValueError(f"Some teams not in layout: {join_and(missing)}.")

    extra = layout_teams - all_teams
    if extra:
        print(f"WARNING: Extra teams in layout will be ignoreed: {join_and(extra)}.")
        for tla in extra:
            ordered_teams.remove(tla)

    return ordered_teams


def order_teams(
    compstate_path: Path,
    team_ids: List[TLA],
    strategy: Strategy,
) -> List[TLA]:
    """
    Order teams either randomly or, if there's a layout available, by location.
    """
    layout_yaml = compstate_path / 'layout.yaml'

    if strategy == Strategy.RANDOM:
        return order_teams_randomly(team_ids)

    elif strategy == Strategy.LAYOUT:
        if not layout_yaml.exists():
            raise ValueError(
                "Unable to order teams by layout when there is no layout file",
            )

        return order_teams_by_location(layout_yaml, team_ids)

    elif strategy == Strategy.AUTO:
        if layout_yaml.exists():
            return order_teams_by_location(layout_yaml, team_ids)

        return order_teams_randomly(team_ids)

    raise AssertionError(
        "Invalid strategy selected. Did you add a new strategy but not add a handler?",
    )



time_parse_pattern = r'^((?P<hours>\d+?)hr)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?$'


class BadDurationException(ValueError):
    def __init__(self, time_str):
        msg = f"Unable to parse duration string '{time_str}'."
        super().__init__(msg)


def parse_duration(time_str):
    import re
    from datetime import timedelta

    parts = re.match(time_parse_pattern, time_str)
    if not parts:
        if not time_str.isdigit():
            raise BadDurationException(time_str)
        else:
            s = int(time_str)
            return timedelta(seconds=s)
    parts = parts.groupdict()
    time_params = {}
    for (name, param) in parts.items():
        if param:
            time_params[name] = int(param)
    return timedelta(**time_params)


def parse_datetime(when_str):
    import re
    from datetime import datetime

    from dateutil.parser import parse as parse_date
    from dateutil.tz import tzlocal

    def parse_now(match):
        return datetime.now()

    def parse_future(match):
        offset = parse_duration(match.group(1))
        return datetime.now() + offset

    def parse_past(match):
        offset = parse_duration(match.group(1))
        return datetime.now() - offset

    def parse_absolute(match):
        return parse_date(match.group(0))

    DATETIME_PATTERNS = [
        (r'^([^ ]+)\s*ago$', parse_past),
        (r'^in\s*([^ ]+)$', parse_future),
        (r'^now$', parse_now),
        (r'^.+$', parse_absolute),
    ]

    for pattern, parse_fn in DATETIME_PATTERNS:
        match = re.match(pattern, when_str)
        if match is None:
            continue
        try:
            when = parse_fn(match)
            break
        except ValueError:
            continue
    else:
        raise ValueError(
            f"Unable to parse date string: {when_str!r}",
        )

    # Timezone information gets ignored, and the resulting datetime is
    # timezone-unaware. However the compstate needs timezone data to be
    # present.
    # Assume that the user wants their current timezone.
    when = when.replace(tzinfo=tzlocal())
    return when


def get_current_match_start(compstate_path):
    from sr.comp.comp import SRComp
    compstate = SRComp(compstate_path)
    now = compstate.schedule.datetime_now
    current_matches = tuple(compstate.schedule.matches_at(now))
    if not current_matches:
        raise Exception("Not currently in a match, specify a valid time instead")

    return min(x.start_time for x in current_matches)


def parse_time(compstate_path, when_str):
    if when_str == "current match":
        return get_current_match_start(compstate_path)
    else:
        return parse_datetime(when_str)


def add_delay(schedule, delay_seconds, when):
    delays = schedule.get('delays')
    if not delays:
        delays = schedule['delays'] = []
    new_delay = {
        'delay': delay_seconds,
        'time': when,
    }
    delays.append(new_delay)


def command(settings):
    import os.path

    from sr.comp.cli import yaml_round_trip as rtyaml

    schedule_path = os.path.join(settings.compstate, "schedule.yaml")
    schedule = rtyaml.load(schedule_path)

    how_long = parse_duration(settings.how_long)
    how_long_seconds = how_long.seconds

    when = parse_time(settings.compstate, settings.when)
    when = when.replace(microsecond=0)

    add_delay(schedule, how_long_seconds, when)

    rtyaml.dump(schedule_path, schedule)

    return how_long, when


def add_arguments(parser):
    parser.add_argument(
        'how_long',
        help=(
            "How long to delay the competition for. Specify either as a number "
            "of seconds or as a string of the form 1m30s."
        ),
    )
    parser.add_argument(
        'when',
        nargs='?',
        default='now',
        help=(
            "When the delay should occur. This can be anything which PHP's "
            "strtotime would be able to parse. Assumes all times are in the "
            "current timezone, regardless of input."
        ),
    )


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'add-delay',
        help="Add a delay the competition state",
    )
    parser.add_argument('compstate', help="competition state repository")
    add_arguments(parser)
    parser.set_defaults(func=command)



def round_name(rounds_left):
    if rounds_left == 0:
        return "Finals"
    elif rounds_left == 1:
        return "Semi Finals"
    elif rounds_left == 2:
        return "Quarter Finals"
    return ""


def command(settings):
    import os.path

    from sr.comp.comp import SRComp

    comp = SRComp(os.path.realpath(settings.compstate))

    teams_last_round = set()
    last_round_num = len(comp.schedule.knockout_rounds) - 1
    for i, matches in enumerate(comp.schedule.knockout_rounds):
        teams_this_round = set()
        for game in matches:
            teams_this_round.update(game.teams)

        print("Teams not in round {} ({})".format(i, round_name(last_round_num - i)))
        out = teams_last_round - teams_this_round
        teams_out = [t for t in out if t is not None]
        for tla in teams_out:
            print(tla, comp.teams[tla].name)
        teams_last_round = teams_this_round
        print()


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'knocked-out-teams',
        help="show the teams knocked out of each knockout round",
    )
    parser.add_argument('compstate', help="competition state repository")
    parser.set_defaults(func=command)



def command(settings):
    from sr.comp.comp import SRComp
    from sr.comp.validation import validate

    comp = SRComp(settings.compstate)

    if settings.lax:
        error_count = 0
    else:
        error_count = validate(comp)

    exit(error_count)


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'validate',
        help="validate the state of a compstate repository",
    )
    parser.add_argument('compstate', help="competition state repository")
    parser.add_argument(
        '-l', '--lax',
        action='store_true',
        help="only check if it loads, rather than run a validation",
    )
    parser.set_defaults(func=command)



def command(args):
    from sr.comp.comp import SRComp

    comp = SRComp(args.compstate)
    matches = comp.schedule.matches

    remaining_teams = dict(comp.teams)

    for slot in matches:
        for arena, match in slot.items():
            print("## Match #{} in Arena {} (at {:%H:%M})".format(
                match.num,
                arena,
                match.start_time,
            ))
            for tla in match.teams:
                team = remaining_teams.get(tla)
                if team:
                    print("  - {tla}: {name}".format(**team._asdict()))
                    remaining_teams.pop(tla)

            if remaining_teams:
                print()
            else:
                return
        if remaining_teams:
            print()
        else:
            return


def add_subparser(subparsers):
    help_msg = "Shows a list of teams, ordered by their first matches."
    description = help_msg + (
        " Output is markdown, and can be converted to PDF by piping through "
        "'pandoc -V geometry:margin=1in'."
    )

    parser = subparsers.add_parser(
        'match-order-teams',
        help=help_msg,
        description=description,
    )
    parser.add_argument('compstate', help="competition state repository")
    parser.set_defaults(func=command)



from setuptools import find_namespace_packages, setup

with open('README.rst') as f:
    long_description = f.read()

with open('setup-requirements.txt') as f:
    setup_requires = f.readlines()

setup(
    name='sr.comp.cli',
    version='1.3.1',
    project_urls={
        'Documentation': 'https://srcomp-cli.readthedocs.org/',
        'Code': 'https://github.com/PeterJCLaw/srcomp-cli',
        'Issue tracker': 'https://github.com/PeterJCLaw/srcomp-cli/issues',
    },
    packages=find_namespace_packages(exclude=('tests',)),
    namespace_packages=['sr', 'sr.comp'],
    description=(
        "Command line tools for interacting with the state of the Student "
        "Robotics Competition"
    ),
    long_description=long_description,
    author="Student Robotics Competition Software SIG",
    author_email='srobo-devel@googlegroups.com',
    install_requires=[
        'python-dateutil >=2.2, <3',
        'paramiko >=1.10, <3',
        'sr.comp >=1.2, <2',
        'reportlab >=3.1.44, <3.6',
        'requests >=2.5.1, <3',
        'ruamel.yaml >=0.13.0, <0.16',
        'simplejson >=3.6, <4',
        'mido >=1.1, <2',
    ],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'srcomp = sr.comp.cli.command_line:main',
        ],
    },
    setup_requires=setup_requires,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
)



"""srcomp command-line interface."""

import argparse
import sys
from typing import List, Optional

from . import (
    add_delay,
    awards,
    delay,
    deploy,
    for_each_match,
    import_schedule,
    knocked_out_teams,
    lighting_controller,
    list_midi_ports,
    match_order_teams,
    print_schedule,
    schedule_league,
    scorer,
    shift_matches,
    show_schedule,
    summary,
    top_match_points,
    update_layout,
    validate,
    yaml_round_trip,
)


def add_list_commands(subparsers: argparse._SubParsersAction) -> None:
    def command(settings):
        commands = subparsers.choices.keys()
        print(" ".join(commands))

    help_text = "Lists the available commands; useful for adding " \
                "auto-completion of command names"

    parser = subparsers.add_parser('list-commands', help=help_text)
    parser.set_defaults(func=command)


def argument_parser() -> argparse.ArgumentParser:
    """A parser for CLI tool command line arguments, from argparse."""
    parser = argparse.ArgumentParser(description="srcomp command-line interface")
    subparsers = parser.add_subparsers(title="commands")
    add_list_commands(subparsers)

    add_delay.add_subparser(subparsers)
    awards.add_subparser(subparsers)
    delay.add_subparser(subparsers)
    deploy.add_subparser(subparsers)
    for_each_match.add_subparser(subparsers)
    import_schedule.add_subparser(subparsers)
    knocked_out_teams.add_subparser(subparsers)
    list_midi_ports.add_subparser(subparsers)
    lighting_controller.add_subparser(subparsers)
    match_order_teams.add_subparser(subparsers)
    print_schedule.add_subparser(subparsers)
    schedule_league.add_subparser(subparsers)
    scorer.add_subparser(subparsers)
    shift_matches.add_subparser(subparsers)
    show_schedule.add_subparser(subparsers)
    summary.add_subparser(subparsers)
    top_match_points.add_subparser(subparsers)
    update_layout.add_subparser(subparsers)
    validate.add_subparser(subparsers)
    yaml_round_trip.add_subparser(subparsers)

    return parser


def main(args: Optional[List[str]] = None) -> None:
    """Run as the CLI tool."""
    if args is None:
        args = sys.argv[1:]
    parser = argument_parser()
    settings = parser.parse_args(args)
    if 'func' in settings:
        settings.func(settings)
    else:
        parser.print_help()



def command(settings):
    import os.path

    from sr.comp.comp import SRComp
    from sr.comp.winners import Award

    comp = SRComp(os.path.realpath(settings.compstate))

    def format_team(tla):
        team = comp.teams[tla]
        return '{} ({}{})'.format(
            tla,
            team.name,
            ' [rookie]' if team.rookie else '',
        )

    award_order = (
        Award.image,
        Award.web,
        Award.committee,
        Award.rookie,
        Award.movement,
        Award.third,
        Award.second,
        Award.first,
    )

    missing = set(Award) - set(award_order)
    assert not missing, "Awards missed!: {}".format(", ".join(map(str, missing)))

    for award in award_order:
        print(f'### {award.value.upper()}')
        recipients = comp.awards.get(award, None)
        if recipients is None:
            print('  Not yet awarded.')
        elif not recipients:
            print('  Awarded to nobody.')
        elif len(recipients) == 1:
            print(" ", format_team(recipients[0]))
        else:
            print('  Split between {} teams (a tie):'.format(len(recipients)))
            for recipient in recipients:
                print("  ", format_team(recipient))
        print()


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'awards',
        help="show who has been given awards",
    )
    parser.add_argument('compstate', help="competition state repo")
    parser.set_defaults(func=command)



_ryaml = None


def _load():
    global _ryaml
    if _ryaml is None:
        import ruamel.yaml as _ryaml

        from sr.comp.yaml_loader import add_time_constructor
        add_time_constructor(_ryaml.RoundTripLoader)
    return _ryaml


def load(yaml_file):
    ryaml = _load()
    with open(yaml_file) as yf:
        return ryaml.load(yf, ryaml.RoundTripLoader, version=(1, 1))


def dump(yaml_file, data):
    ryaml = _load()
    yaml = ryaml.dump(data, Dumper=ryaml.RoundTripDumper)
    yaml = "\n".join(x.rstrip() for x in yaml.splitlines())
    with open(yaml_file, 'w') as yf:
        print(yaml, file=yf)


def command(settings):
    fp = settings.file_path
    dump(fp, load(fp))


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'round-trip',
        help="Round-trip a yaml file using compstate loading",
    )
    parser.add_argument('file_path', help="target file to round trip")
    parser.set_defaults(func=command)
